@@ -165,26 +165,20 @@
 #![warn(rust_2018_idioms)]
 #![cfg_attr(docsrs, feature(doc_cfg))]
 #![no_std]
-#![deprecated = "this crate is unsound because of incorrect trait bounds \
-                 regarding thread safety. please migrate to cryo 0.3"]
 
 use core::{
-    cell::UnsafeCell,
     fmt,
     marker::{PhantomData, PhantomPinned},
     ops::{Deref, DerefMut},
     pin::Pin,
     ptr::NonNull,
 };
+use pin_utils::pin_mut;
 use stable_deref_trait::{CloneStableDeref, StableDeref};
 
 #[cfg(feature = "std")]
 extern crate std;
 
-// Used by `cryo!`
-#[doc(hidden)]
-pub use pin_utils::pin_mut;
-
 mod lock;
 pub use self::lock::*;
 
@@ -202,16 +196,42 @@ pub use self::lock::*;
 ///
 /// [module-level documentation]: index.html
 pub struct Cryo<'a, T: ?Sized, Lock: crate::Lock> {
-    state: UnsafeCell<State<T, Lock>>,
-    _phantom: (PhantomData<&'a T>, PhantomPinned),
+    state: State<T, Lock>,
+    _phantom: PhantomData<&'a T>,
 }
 
-unsafe impl<'a, T: ?Sized + Send, Lock: crate::Lock> Send for Cryo<'a, T, Lock> where
-    Lock::LockMarker: Send
+/// `Cryo` may be moved around multiple threads, and on each thread
+/// [`CryoRef`] may be created, forming multiple instances of `&T`.
+/// Therefore `CryoMut: Send` necessitates `T: Sync`.
+///
+/// Another interpretation is to think of `Cryo` as `&T` with an erased
+/// lifetime.
+///
+/// `T: Send` is not necessary because `Cryo` doesn't provide `&mut T`.
+///
+/// It owns `Lock`, hence the `Lock: Send` requirement. Furthermore,
+/// `Lock` can be shared with other threads through [`CryoMutReadGuard`], hence
+/// `Lock: Sync`.
+///
+/// `Lock::LockMarker: Send` is required because `Lock::lock_` may be called
+/// after `Cryo` is sent to another thread.
+unsafe impl<'a, T: ?Sized + Sync, Lock: crate::Lock> Send for Cryo<'a, T, Lock>
+where
+    Lock: Send + Sync,
+    Lock::LockMarker: Send,
 {
 }
-unsafe impl<'a, T: ?Sized + Send + Sync, Lock: crate::Lock> Sync for Cryo<'a, T, Lock> where
-    Lock::LockMarker: Send
+
+/// `&T` can be created from `&Cryo`, so `Cryo: Sync` necessitates `T: Sync`.
+///
+/// `T: Send` is not necessary because `Cryo` doesn't provide `&mut T`.
+///
+/// `Lock::LockMarker: Send` is required because `Lock::lock_` may be called
+/// through an `&Cryo` sent to another thread.
+unsafe impl<'a, T: ?Sized + Sync, Lock: crate::Lock> Sync for Cryo<'a, T, Lock>
+where
+    Lock: Sync,
+    Lock::LockMarker: Send,
 {
 }
 
@@ -229,22 +249,43 @@ unsafe impl<'a, T: ?Sized + Send + Sync, Lock: crate::Lock> Sync for Cryo<'a, T,
 ///
 /// [module-level documentation]: index.html
 pub struct CryoMut<'a, T: ?Sized, Lock: crate::Lock> {
-    state: UnsafeCell<State<T, Lock>>,
+    state: State<T, Lock>,
     _phantom: (PhantomData<&'a mut T>, PhantomPinned),
 }
 
-unsafe impl<'a, T: ?Sized + Send, Lock: crate::Lock> Send for CryoMut<'a, T, Lock> where
-    Lock::LockMarker: Send
+/// `CryoMut` may be moved around multiple threads, and on each thread
+/// [`CryoMutReadGuard`] may be created, forming multiple instances of `&T`.
+/// Therefore `CryoMut: Send` necessitates `T: Sync`.
+///
+/// It owns `Lock`, hence the `Lock: Send` requirement. Furthermore,
+/// `Lock` can be shared with other threads through [`CryoMutReadGuard`], hence
+/// `Lock: Sync`.
+///
+/// `Lock::LockMarker: Send` is required because `Lock::lock_` may be called
+/// after `Cryo` is sent to another thread.
+unsafe impl<'a, T: ?Sized + Send + Sync, Lock: crate::Lock> Send for CryoMut<'a, T, Lock>
+where
+    Lock: Send + Sync,
+    Lock::LockMarker: Send,
 {
 }
-unsafe impl<'a, T: ?Sized + Send + Sync, Lock: crate::Lock> Sync for CryoMut<'a, T, Lock> where
-    Lock::LockMarker: Send
+
+/// `&mut T` may be created from `&CryoMut`, so sending `&CryoMut` to another
+/// thread requires `T: Send`.
+///
+/// `Lock::LockMarker: Send` is required because `Lock::lock_` may be called
+/// through an `&Cryo` sent to another thread.
+unsafe impl<'a, T: ?Sized + Send + Sync, Lock: crate::Lock> Sync for CryoMut<'a, T, Lock>
+where
+    Lock: Sync,
+    Lock::LockMarker: Send,
 {
 }
 
 struct State<T: ?Sized, Lock> {
     data: NonNull<T>,
     lock: Lock,
+    _phantom: PhantomPinned,
 }
 
 /// The lock guard type of [`Cryo`]. This is currently a type alias but might
@@ -256,25 +297,41 @@ pub struct CryoMutReadGuard<T: ?Sized, Lock: crate::Lock> {
     state: NonNull<State<T, Lock>>,
 }
 
-unsafe impl<T: ?Sized + Send, Lock: crate::Lock> Send for CryoMutReadGuard<T, Lock> where
-    Lock::UnlockMarker: Send
-{
-}
-unsafe impl<T: ?Sized + Send + Sync, Lock: crate::Lock> Sync for CryoMutReadGuard<T, Lock> where
-    Lock::UnlockMarker: Send
+/// `CryoMutReadGuard` is essentially `&T` with an indeterminate lifetime.
+/// The owning thread may be constrained by [`Lock::UnlockMarker`].
+unsafe impl<T: ?Sized + Sync, Lock: crate::Lock> Send for CryoMutReadGuard<T, Lock>
+where
+    Lock: Sync,
+    Lock::UnlockMarker: Send,
 {
 }
 
+/// `CryoMutReadGuard` is essentially `&T` with an indeterminate lifetime.
+///
+/// `Lock` is never touched through `&CryoMutReadGuard<_, Lock>`, so the trait
+/// bounds is not constrained by `Lock`.
+unsafe impl<T: ?Sized + Sync, Lock: crate::Lock> Sync for CryoMutReadGuard<T, Lock> {}
+
 /// The write lock guard type of [`CryoMut`].
 pub struct CryoMutWriteGuard<T: ?Sized, Lock: crate::Lock> {
     state: NonNull<State<T, Lock>>,
 }
 
-unsafe impl<T: ?Sized + Send, Lock: crate::Lock> Send for CryoMutWriteGuard<T, Lock> where
-    Lock::UnlockMarker: Send
+/// `CryoMutWriteGuard` is essentially `&mut T` with an indeterminate lifetime.
+/// The owning thread may be constrained by [`Lock::UnlockMarker`].
+unsafe impl<T: ?Sized + Send, Lock: crate::Lock> Send for CryoMutWriteGuard<T, Lock>
+where
+    Lock: Sync,
+    Lock::UnlockMarker: Send,
 {
 }
 
+/// `CryoMutWriteGuard` is essentially `&mut T` with an indeterminate lifetime.
+///
+/// `Lock` is never touched through `&CryoMutWriteGuard<_, Lock>`, so the trait
+/// bounds is not constrained by `Lock`.
+unsafe impl<T: ?Sized + Sync, Lock: crate::Lock> Sync for CryoMutWriteGuard<T, Lock> {}
+
 impl<'a, T: ?Sized + 'a, Lock: crate::Lock> Cryo<'a, T, Lock> {
     /// Construct a new `Cryo`.
     ///
@@ -315,21 +372,22 @@ impl<'a, T: ?Sized + 'a, Lock: crate::Lock> Cryo<'a, T, Lock> {
     #[inline]
     pub unsafe fn new(x: &'a T) -> Self {
         Self {
-            state: UnsafeCell::new(State {
+            state: State {
                 data: NonNull::from(x),
                 lock: Lock::new(),
-            }),
-            _phantom: (PhantomData, PhantomPinned),
+                _phantom: PhantomPinned,
+            },
+            _phantom: PhantomData,
         }
     }
 
     /// Borrow a cell using runtime lifetime rules.
     #[inline]
     pub fn borrow(self: Pin<&Self>) -> CryoRef<T, Lock> {
-        // Safety: `Cryo`'s `Send`-ness is constrained by that of `Lock::LockMarker`
-        unsafe { (*self.state.get()).lock.lock_shared() };
+        // Safety: `&Cryo`'s `Send`-ness is constrained by that of `Lock::LockMarker`
+        unsafe { self.state.lock.lock_shared() };
         CryoRef {
-            state: NonNull::new(self.state.get()).unwrap(),
+            state: NonNull::from(&self.state),
         }
     }
 
@@ -338,7 +396,7 @@ impl<'a, T: ?Sized + 'a, Lock: crate::Lock> Cryo<'a, T, Lock> {
     /// This operation is no-op since `Cryo` only can be immutably borrowed.
     #[inline]
     pub fn get(&self) -> &'a T {
-        unsafe { &*(*self.state.get()).data.as_ptr() }
+        unsafe { &*self.state.data.as_ptr() }
     }
 }
 
@@ -351,8 +409,8 @@ impl<'a, T: ?Sized + fmt::Debug, Lock: crate::Lock> fmt::Debug for Cryo<'a, T, L
 impl<'a, T: ?Sized + 'a, Lock: crate::Lock> Drop for Cryo<'a, T, Lock> {
     #[inline]
     fn drop(&mut self) {
-        // Safety: `Cryo`'s `Send`-ness is constrained by that of `Lock::LockMarker`
-        unsafe { (*self.state.get()).lock.lock_exclusive() };
+        // Safety: `&Cryo`'s `Send`-ness is constrained by that of `Lock::LockMarker`
+        unsafe { self.state.lock.lock_exclusive() };
         // A write lock ensures there are no other references to
         // the contents
     }
@@ -371,10 +429,11 @@ impl<'a, T: ?Sized + 'a, Lock: crate::Lock> CryoMut<'a, T, Lock> {
     #[inline]
     pub unsafe fn new(x: &'a mut T) -> Self {
         Self {
-            state: UnsafeCell::new(State {
+            state: State {
                 data: NonNull::from(x),
                 lock: Lock::new(),
-            }),
+                _phantom: PhantomPinned,
+            },
             _phantom: (PhantomData, PhantomPinned),
         }
     }
@@ -382,20 +441,20 @@ impl<'a, T: ?Sized + 'a, Lock: crate::Lock> CryoMut<'a, T, Lock> {
     /// Acquire a read (shared) lock on a `CryoMut`.
     #[inline]
     pub fn read(self: Pin<&Self>) -> CryoMutReadGuard<T, Lock> {
-        // Safety: `CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
-        unsafe { (*self.state.get()).lock.lock_shared() };
+        // Safety: `&CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
+        unsafe { self.state.lock.lock_shared() };
         CryoMutReadGuard {
-            state: NonNull::new(self.state.get()).unwrap(),
+            state: NonNull::from(&self.state),
         }
     }
 
     /// Attempt to acquire a read (shared) lock on a `CryoMut`.
     #[inline]
     pub fn try_read(self: Pin<&Self>) -> Option<CryoMutReadGuard<T, Lock>> {
-        // Safety: `CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
-        if unsafe { (*self.state.get()).lock.try_lock_shared() } {
+        // Safety: `&CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
+        if unsafe { self.state.lock.try_lock_shared() } {
             Some(CryoMutReadGuard {
-                state: NonNull::new(self.state.get()).unwrap(),
+                state: NonNull::from(&self.state),
             })
         } else {
             None
@@ -405,20 +464,20 @@ impl<'a, T: ?Sized + 'a, Lock: crate::Lock> CryoMut<'a, T, Lock> {
     /// Acquire a write (exclusive) lock on a `CryoMut`.
     #[inline]
     pub fn write(self: Pin<&Self>) -> CryoMutWriteGuard<T, Lock> {
-        // Safety: `CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
-        unsafe { (*self.state.get()).lock.lock_exclusive() };
+        // Safety: `&CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
+        unsafe { self.state.lock.lock_exclusive() };
         CryoMutWriteGuard {
-            state: NonNull::new(self.state.get()).unwrap(),
+            state: NonNull::from(&self.state),
         }
     }
 
     /// Attempt to acquire a write (exclusive) lock on a `CryoMut`.
     #[inline]
     pub fn try_write(self: Pin<&Self>) -> Option<CryoMutWriteGuard<T, Lock>> {
-        // Safety: `CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
-        if unsafe { (*self.state.get()).lock.try_lock_exclusive() } {
+        // Safety: `&CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
+        if unsafe { self.state.lock.try_lock_exclusive() } {
             Some(CryoMutWriteGuard {
-                state: NonNull::new(self.state.get()).unwrap(),
+                state: NonNull::from(&self.state),
             })
         } else {
             None
@@ -434,7 +493,7 @@ impl<'a, T: ?Sized + 'a, Lock: crate::Lock> CryoMut<'a, T, Lock> {
         // FIXME: The lifetime elision is not possible here because of
         //        <https://github.com/rust-lang/rust/issues/52675>
         if self.as_ref().try_write().is_some() {
-            Some(unsafe { &mut *(*self.state.get()).data.as_ptr() })
+            Some(unsafe { &mut *self.state.data.as_ptr() })
         } else {
             None
         }
@@ -465,8 +524,8 @@ impl<'a, T: ?Sized + fmt::Debug, Lock: crate::Lock> fmt::Debug for CryoMut<'a, T
 impl<'a, T: ?Sized + 'a, Lock: crate::Lock> Drop for CryoMut<'a, T, Lock> {
     #[inline]
     fn drop(&mut self) {
-        // Safety: `CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
-        unsafe { (*self.state.get()).lock.lock_exclusive() };
+        // Safety: `&CryoMut`'s `Send`-ness is constrained by that of `Lock::LockMarker`
+        unsafe { self.state.lock.lock_exclusive() };
         // A write lock ensures there are no other references to
         // the contents
     }
@@ -562,140 +621,6 @@ impl<T: ?Sized, Lock: crate::Lock> Drop for CryoMutWriteGuard<T, Lock> {
     }
 }
 
-/// Construct a [`Cryo`] or [`CryoMut`] and bind it to a local variable.
-///
-/// # Safety
-///
-/// **Don't use. This macro is unsound when used inside an `async fn`.** This
-/// macro doesn't require `unsafe { ... }` merely not to cause breakage.
-///
-/// The unsafety is demonstrated in the following code:
-///
-/// ```should_panic
-/// #![allow(deprecated)]
-/// use cryo::cryo;
-/// use std::{
-///     sync::atomic::{AtomicBool, Ordering},
-///     future::Future,
-///     task::Context,
-/// };
-///
-/// // For demonstration purposes, we want to stop execution when an undefined
-/// // behavior is about occur. To this end, we track the `UserType` object's
-/// // aliveness with this flag.
-/// static IS_USER_TYPE_ALIVE: AtomicBool = AtomicBool::new(true);
-///
-/// struct UserType(u32);
-///
-/// impl Drop for UserType {
-///     fn drop(&mut self) {
-///         IS_USER_TYPE_ALIVE.store(false, Ordering::Relaxed);
-///     }
-/// }
-///
-/// // Let there be a `UserType`.
-/// let user_type = UserType(42);
-///
-/// let mut borrow = None;
-///
-/// // Apply `cryo!` on it inside an `async` block.
-/// let mut fut = Box::pin(async {
-///     cryo!(let cryo: Cryo<_, cryo::SyncLock> = &user_type);
-///
-///     // Leak `borrow` to the outer environment
-///     borrow = Some(cryo.borrow());
-///
-///     // This `Future` will get stuck here. Furthermore, we `forget` this
-///     // `Future`, so `cryo`'s destructor will never run.
-///     std::future::pending::<()>().await
-/// });
-///
-/// // Run the `Future` until it stalls
-/// fut.as_mut().poll(&mut Context::from_waker(&futures::task::noop_waker()));
-///
-/// // Forget the `Future`. The compiler thinks `user_type` is not borrowed,
-/// // but in fact `cryo`, which is borrowing it, is still on memory.
-/// std::mem::forget(fut);
-///
-/// // And `user_type` is gone. Now `cryo` is dangling.
-/// drop(user_type);
-///
-/// // But we can still access the dead `user_type` through `borrow`!
-/// let borrow = borrow.unwrap();
-/// assert!(
-///     IS_USER_TYPE_ALIVE.load(Ordering::Relaxed),
-///     "`cryo!` is supposed to keep us safe, isn't it?"  // well, it betrayed us. (panics)
-/// );
-/// assert_eq!(borrow.0, 42);  // UB
-/// ```
-///
-/// # Examples
-///
-/// ```
-/// #![allow(deprecated)]
-/// use cryo::cryo;
-/// cryo!(let cryo: Cryo<u8> = &42);
-/// assert_eq!(*cryo.borrow(), 42);
-/// ```
-///
-/// ```
-/// #![allow(deprecated)]
-/// use cryo::cryo;
-/// let mut var = 42;
-/// {
-///     cryo!(let cryo: CryoMut<u8> = &mut var);
-///     *cryo.write() = 84;
-/// }
-/// assert_eq!(var, 84);
-/// ```
-///
-/// The lock implementation can be specified by an extra generic argument. It
-/// defaults to [`LocalLock`] when unspecified.
-///
-/// ```
-/// #![allow(deprecated)]
-/// use cryo::cryo;
-/// use std::thread::spawn;
-/// cryo!(let cryo: Cryo<_, cryo::SyncLock> = &42);
-/// let borrow = cryo.borrow();
-/// spawn(move || {
-///     assert_eq!(*borrow, 42);
-/// });
-/// ```
-#[deprecated = "`cryo!` is unsound when used inside `async fn` and will be \
-                removed in a future version"]
-#[macro_export]
-macro_rules! cryo {
-    // empty (base case for the recursion)
-    () => {};
-
-    // process multiple declarations
-    ($(#[$attr:meta])* let $name:ident: $Cryo:ident< $t:ty $(, $Lock:ty)? > = $init:expr; $($rest:tt)*) => (
-        $crate::cryo!($(#[$attr])* let $name: $Cryo<$t $(, $Lock)?> = $init);
-        $crate::cryo!($($rest)*);
-    );
-
-    // handle a single declaration
-    ($(#[$attr:meta])* let $name:ident: $Cryo:ident< $t:ty $(, $Lock:ty)? > = $init:expr) => (
-        let cryo = unsafe { $crate::$Cryo::<$t, $crate::__LockOrDefault!($(($Lock))?)>::new($init) };
-        $crate::pin_mut!(cryo);
-        let $name = cryo.into_ref();
-    );
-}
-
-#[doc(hidden)]
-#[macro_export]
-macro_rules! __LockOrDefault {
-    // Custom
-    (($t:ty)) => {
-        $t
-    };
-    // Default
-    () => {
-        $crate::LocalLock
-    };
-}
-
 /// The trait for types that can be wrapped with [`Cryo`] or [`CryoMut`].
 pub trait WithCryo: private::Sealed + Sized {
     type Cryo;
